#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <sys/ioctl.h>
#include <asm/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nftl.h>



char readbuf[512];

struct NFTLMediaHeader MedHead[2];
int NumMedHeads = 0;

unsigned char BadBlockTable[8192 - 512];

#define ERASESIZE 0x2000
#define NUMVUNITS ((40*1024*1024) / ERASESIZE)

unsigned short EUNtable[NUMVUNITS];
union nftl_uci UCItable[NUMVUNITS][3];

unsigned short nextEUN(unsigned short curEUN)
{
  return UCItable[curEUN][0].a.ReplUnitNum;
}

void main(int argc, char **argv)
{
  unsigned long ofs;
  int i,j;
  int fd = open("/dev/mtd0", O_RDONLY);
  int ofd = -1;
  unsigned char databuf[512];
  struct nftl_oob oobbuf;
  struct mtd_oob_buf  oob = {0,16,(unsigned char *)&oobbuf};

  if (fd == -1) {
    perror("open flash");
    exit (1);
  }
  
  if (argc > 1) {
    ofd = open(argv[1], O_WRONLY | O_TRUNC | O_CREAT, 0644);
    if (ofd == -1)
      perror ("open outfile");
  }
    
  for (ofs = 0; ofs < 4 * 1024 * 1024 ; ofs += ERASESIZE)
    {
      pread (fd, &MedHead[NumMedHeads], sizeof(struct NFTLMediaHeader), ofs);
      
      if (!strncmp(MedHead[NumMedHeads].DataOrgID, "ANAND", 6)) {
	if (!NumMedHeads) {
	  printf("NFTL Media Header found at offset %lx:\n", ofs);
	  printf("NumEraseUnits:    %d\n",MedHead[NumMedHeads].NumEraseUnits);
	  printf("FirstPhysicalEUN: %d\n",MedHead[NumMedHeads].FirstPhysicalEUN);
	  printf("Formatted Size:   %d\n",MedHead[NumMedHeads].FormattedSize);
	  printf("UnitSizeFactor:   0x%x\n",MedHead[NumMedHeads].UnitSizeFactor);
	  pread(fd, BadBlockTable, 8192-512, ofs + 512);
	}
	else
	  printf("Second NFTL Media Header found at offset %lx\n",ofs);

	NumMedHeads ++;
      }
    }
  if (NumMedHeads == 0) {
    printf("No NFTL found on device\n");
    exit(1);
  }
  memset(EUNtable, 0, sizeof(EUNtable));
 
  for (i=MedHead[0].FirstPhysicalEUN; i < MedHead[0].FirstPhysicalEUN + 
	 MedHead[0].NumEraseUnits ; i++) {
    /* For each EUN */
    
    ofs = i * ERASESIZE;
    
    for (j=0; j<3; j++) {
      oob.start = ofs + (j*512);
      ioctl(fd, MEMREADOOB, &oob);
      memcpy(&UCItable[i][j], &oobbuf.u, 8);
    }
    
    if (UCItable[i][1].b.EraseMark != 0x3c69)
      printf("EraseMark not present in unit %d: %x\n",i, UCItable[i][1].b.EraseMark);
    else
      //      printf("EUN #%d is Virtual Unit %d, replaced by %d\n", i, 
      //	     UCItable[i][0].a.VirtUnitNum & 0x7fff,
      //     UCItable[i][0].a.ReplUnitNum != 0xffff?UCItable[i][0].a.ReplUnitNum:0);

      if (!(UCItable[i][0].a.VirtUnitNum & 0x8000)) {
	/* If this is the first in a chain, store the EUN */
	//	printf("Setting EUNtable[%d] to %d\n", 
	//     UCItable[i][0].a.VirtUnitNum & 0x7fff, i);
	      if (EUNtable[UCItable[i][0].a.VirtUnitNum & 0x7fff]) {
		      printf("Duplicate start of chain for VUC %d: Unit %d replaces Unit %d\n",
			     UCItable[i][0].a.VirtUnitNum & 0x7fff, i,
			     EUNtable[UCItable[i][0].a.VirtUnitNum & 0x7fff]    );
	      }
	EUNtable[UCItable[i][0].a.VirtUnitNum & 0x7fff] = i;
      }
    if ( UCItable[i][0].a.VirtUnitNum == 0xffff)
	    printf("Unit %d is free\n", i);
    else
	    printf("Unit %d is in chain %d and %s a replacement\n", i, UCItable[i][0].a.VirtUnitNum & 0x7fff,
		   UCItable[i][0].a.VirtUnitNum & 0x8000?"is":"is not");
  } 

  for (i=0; i< /* NUMVUNITS */ (MedHead[0].FormattedSize / 0x2000) + 1; i++) {
      unsigned short curEUN = EUNtable[i];

      printf("Virtual Unit #%d: ", i);
      if (!curEUN) {
	printf("Not present\n");
	continue;
      }
      printf("%d", curEUN);
            
      while ((curEUN = nextEUN(curEUN)) != 0xffff) {
	printf(", %d", curEUN & 0x7fff);
      }
      printf("\n");
    
      if (ofd != -1) {
	/* Actually write out the data */
	for (j=0; j < ERASESIZE / 512; j++) {
	  /* For each sector in the block */
	  unsigned short lastgoodEUN = 0xffff, thisEUN = EUNtable[i];
	  
	  if (thisEUN == 0xffff) thisEUN = 0;
	  
	  while (thisEUN && (thisEUN & 0x7fff) != 0x7fff) {
	    oob.start = (thisEUN * ERASESIZE) + (j * 512);
	    ioctl(fd, MEMREADOOB, &oob);
	    
	    
	    switch (oobbuf.b.Status) {
	    case 0xffff: /* This is still free. Don't look any more */
	      thisEUN = 0;
	      break;
	      
	    case 0x5555: /* BLOCK_USED. This is a good one. */
	      lastgoodEUN = thisEUN;
	      break;
	    }
	    
	    /* Find the next erase unit in this chain, if any */
	    if (thisEUN)
	      thisEUN = nextEUN(thisEUN) & 0x7fff;
	  }
	  if (lastgoodEUN == 0xffff)
	    memset(readbuf,0,512);
	  else
	    pread(fd, readbuf, 512, (lastgoodEUN * ERASESIZE) + (j * 512));
	  
	  write(ofd, readbuf, 512);
	}
	
      }
    }
}
